in_vim = True

from ribosome.nvim import NvimFacade
from ribosome.plugin import NvimStatePlugin, NvimPlugin, AutoPlugin
from ribosome.request.command import command, msg_command, json_msg_command
from ribosome.request.function import function, msg_function, json_msg_function

__all__ = ('command', 'NvimFacade', 'NvimStatePlugin', 'msg_command', 'NvimPlugin', 'in_vim', 'json_msg_command',
           'function', 'msg_function', 'json_msg_function', 'AutoPlugin')



import functools
from typing import TypeVar, Type, Callable, Tuple, Generator, Any
from types import ModuleType

from neovim.msgpack_rpc import MsgpackStream, AsyncSession, Session
from neovim.plugin import Host
from neovim.api import Nvim
from neovim.msgpack_rpc.event_loop.base import BaseEventLoop
from neovim.msgpack_rpc.event_loop.asyncio import AsyncioEventLoop
from neovim.msgpack_rpc.event_loop.uv import UvEventLoop

from amino import Either, _, L, Maybe, Lists, amino_log, Logger, __, Path
from amino.either import ImportFailure
from amino.func import Val
from amino.logging import amino_root_file_logging
from amino.do import tdo

from ribosome.nvim import NvimFacade
from ribosome.logging import ribo_log, Logging
from ribosome.rpc import rpc_handler_functions, RpcHandlerFunction, define_handlers
from ribosome import NvimPlugin, AutoPlugin
from ribosome.settings import Config
from ribosome import options

Loop = TypeVar('Loop', bound=BaseEventLoop)
NP = TypeVar('NP', bound=NvimPlugin)
D = TypeVar('D')
AS = TypeVar('AS')
A = TypeVar('A', bound=AS)
C = TypeVar('C', bound=Config)


class PluginHost(Host, Logging):

    def _load(self, data: Tuple[str, Type[NP]]) -> None:
        amino_log.debug(f'loading host with {data}')
        file, tpe = data
        name = tpe.name or 'ribosome'
        vim = NvimFacade(self._configure_nvim_for(tpe), name)
        instance = tpe(vim.vim)
        handlers = rpc_handler_functions(instance)
        specs = handlers / _.spec
        handlers % L(self.register)(file, _)
        self._specs[file] = specs / _.encode
        define_handlers(vim.channel_id, specs, name, file).attempt(vim).get_or_raise
        amino_log.debug(f'defined handlers for {file}')

    def register(self, file: str, handler: RpcHandlerFunction) -> None:
        spec = handler.spec
        method = spec.rpc_method(file)
        func = functools.partial(self._wrap_function, handler.func, spec.sync, True, None, method)
        self._copy_attributes(handler.func, func)
        target = self._request_handlers if spec.sync else self._notification_handlers
        target[method] = func
        amino_log.debug(f'registered {handler} for {file}')


def session(Loop: Loop, *args: str, transport_type: str='stdio', **kwargs: str) -> Session:
    return Session(AsyncSession(MsgpackStream(Loop(transport_type, *args, **kwargs))))


def host(loop: Loop) -> PluginHost:
    sess = session(loop)
    nvim = Nvim.from_session(sess)
    return PluginHost(nvim)


def start_host(prefix: str, tpe: Type[NP], uv_loop: bool) -> int:
    Loop = UvEventLoop if uv_loop else AsyncioEventLoop
    host(Loop).start((prefix, tpe))
    return 0


@tdo(Either[str, A])
def instance_from_module(mod: ModuleType, pred: Callable[[Any], bool], desc: str) -> Generator:
    all = yield Maybe.getattr(mod, '__all__').to_either(f'module `{mod.__name__}` does not define `__all__`')
    yield (
        Lists.wrap(all)
        .flat_map(L(Maybe.getattr)(mod, _))
        .find(pred)
        .to_either(f'no {desc} in ``{mod.__name__}.__all__`')
    )


def cls_from_module(mod: ModuleType, tpe: Type[AS]) -> Either[str, Type[A]]:
    pred = lambda a: isinstance(a, type) and issubclass(a, tpe)
    return instance_from_module(mod, pred, f'subclass of `{tpe}`')


def plugin_cls_from_module(mod: ModuleType) -> Either[str, Type[NP]]:
    return cls_from_module(mod, NvimPlugin)


def config_from_module(mod: ModuleType) -> Either[str, Type[C]]:
    pred = lambda a: isinstance(a, Config)
    return instance_from_module(mod, pred, 'instance of `Config`')


log_initialized = False


def nvim_log() -> Logger:
    global log_initialized
    if not log_initialized:
        NvimFacade.stdio_with_logging('ribosome_start_host')
        log_initialized = True
    return ribo_log


def start_config_stage_2(cls: Either[str, Type[NP]], config: Config) -> int:
    sup = cls | Val(AutoPlugin)
    debug = options.development.exists
    class Plug(sup, config=config, pname=config.name, prefix=config.prefix, debug=debug):
        def __init__(self, vim: Nvim) -> None:
            super().__init__(vim, config)
    amino_log.debug(f'starting plugin from {config} with superclass {sup}, debug: {debug}')
    return start_host(config.name, Plug, True)


def error(msg: str) -> int:
    amino_log.error(msg)
    nvim_log().error(msg)
    return 1


def import_error(e: ImportFailure, desc: str) -> int:
    return error(e.expand.join_lines)


def exception(e: Exception, desc: str) -> int:
    f = __.caught_exception_error(f'starting host from {desc}', e)
    f(amino_log)
    f(nvim_log())
    return 1


def start_config_stage_1(mod: ModuleType) -> int:
    plugin_cls = plugin_cls_from_module(mod)
    return config_from_module(mod).cata(error, L(start_config_stage_2)(plugin_cls, _))


def setup_log() -> None:
    amino_root_file_logging()


def start_from(source: str, importer: Callable[[str], Either[ImportFailure, ModuleType]], desc: str) -> int:
    try:
        setup_log()
        amino_log.debug(f'start_{desc}: {source}')
        return importer(source).cata(L(import_error)(_, source), start_config_stage_1)
    except Exception as e:
        return exception(e, source)


def start_module(mod: str) -> int:
    return start_from(mod, Either.import_module, 'module')


def start_file(path: str) -> int:
    p = Path(path)
    file = p / '__init__.py' if p.is_dir() else p
    return start_from(str(file), Either.import_file, 'file')


__all__ = ('start_host', 'start_module', 'start_file')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0x353c2c54

# Compiled with Coconut version 1.3.0 [Dead Parrot]

# Coconut Header: -------------------------------------------------------------

from __future__ import generator_stop
import sys as _coconut_sys, os.path as _coconut_os_path
_coconut_file_path = _coconut_os_path.dirname(_coconut_os_path.abspath(__file__))
_coconut_sys.path.insert(0, _coconut_file_path)
from __coconut__ import _coconut, _coconut_NamedTuple, _coconut_MatchError, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_pipe, _coconut_star_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial
from __coconut__ import *
_coconut_sys.path.remove(_coconut_file_path)

# Compiled Coconut: -----------------------------------------------------------

from typing import TypeVar  # line 1

from lenses import Lens  # line 3

from amino.do import tdo  # line 5
from amino import IO  # line 6
from amino import List  # line 6
from amino import Path  # line 6

from ribosome.nvim import NvimIO  # line 8
from ribosome.nvim.io import NvimIOState  # line 9
from ribosome.record import decode_json  # line 10
from ribosome.record import encode_json  # line 10
from ribosome.settings import AutoData  # line 11

A = TypeVar('A')  # line 13
D = TypeVar('D', bound=AutoData)  # line 14


@tdo(NvimIOState[D, Path])  # line 17
def history_file(name: 'str') -> 'Generator':  # line 18
    settings = yield NvimIOState.inspect(lambda a: a.settings)  # line 19
    dir = yield NvimIOState.lift(settings.state_dir.value_or_default)  # line 20
    yield NvimIOState.lift(NvimIO.from_io(IO.delay(dir.mkdir, parents=True, exist_ok=True)))  # line 21
    yield NvimIOState.pure(dir / f'{name}.json')  # line 22


@tdo(NvimIOState[D, None])  # line 25
def load_json_state(name: 'str', l: 'Lens') -> 'Generator':  # line 26
    file = yield history_file(name)  # line 27
    exists = yield NvimIOState.lift(NvimIO.from_io(IO.delay(file.exists)))  # line 28
    if exists:  # line 29
        json = yield NvimIOState.lift(NvimIO.from_io(IO.delay(file.read_text)))  # line 30
        data = yield NvimIOState.lift(NvimIO.from_either(decode_json(json)))  # line 31
        yield NvimIOState.modify(lambda a: l.bind(a).set(data))  # line 32
    else:  # line 33
        yield NvimIOState.pure(None)  # line 34

@tdo(NvimIOState[D, None])  # line 36
def store_json_state(name: 'str', data: '_coconut.typing.Callable[[D], A]') -> 'Generator':  # line 37
    file = yield history_file(name)  # line 38
    payload = yield NvimIOState.inspect(data)  # line 39
    json = yield NvimIOState.lift(NvimIO.from_either(encode_json(payload)))  # line 40
    yield NvimIOState.lift(NvimIO.from_io(IO.delay(file.write_text, json)))  # line 41
    yield NvimIOState.pure(None)  # line 42

__all__ = ('load_json_state', 'store_json_state')  # line 44



import abc
from typing import TypeVar

from ribosome.logging import Logging
from ribosome.machine.message_base import Message

from amino import Maybe, _, __


M = TypeVar('M', bound=Message)


class Machine(Logging, abc.ABC):

    @abc.abstractproperty
    def parent(self) -> Maybe['Machine']:
        ...

    @abc.abstractproperty
    def name(self) -> str:
        ...

    def bubble(self, msg: M) -> None:
        self.parent.cata(_.bubble, lambda: self.send)(msg)

    def log_message(self, msg: Message, name: str) -> None:
        return self.parent % __.log_message(msg, name)


__all__ = ('Machine',)



import re
import typing
from typing import TypeVar, Callable, Any, Union, Optional, Generic, Type
from pathlib import Path
import threading
from concurrent import futures
from contextlib import contextmanager
import asyncio
import abc
from types import FunctionType
import inspect
import traceback
import time
import functools

import neovim
from neovim.api import NvimError, Nvim
from neovim.api.common import Remote

from amino.util.string import camelcaseify

import amino
from amino import Maybe, List, Map, Boolean, Empty, Just, __, _, Try, Either, Left, Right, L, Lists
from amino.util.string import decode
from amino.util.fun import format_funcall
from amino.lazy import lazy
from amino.io import IOException
from amino.util.numeric import parse_int

import ribosome
from ribosome.logging import Logging


def squote(text: str) -> str:
    return text.replace("'", "''")


def dquote(text: str) -> str:
    return text.replace('"', '\\"')


def quote(text: str) -> str:
    return dquote(squote(text))


A = TypeVar('A')


def echo(text: Union[str, List[str]], cmd: str='echom', prefix: Maybe[str]=Empty()) -> List[str]:
    lines = text if isinstance(text, List) else Lists.lines(str(text))
    pre = prefix.map(_ + ': ') | ''
    return lines.map(lambda a: '{} "{}{}"'.format(cmd, pre, dquote(a)))


def echohl(text: Union[str, List[str]], hl: str, prefix: Maybe[str]=Empty()) -> List[str]:
    return echo(text, prefix=prefix).cons(f'echohl {hl}').cat('echohl None')


def not_on_main_thread() -> bool:
    return threading.current_thread() != threading.main_thread()


shutdown = False

R = TypeVar('R', bound=Remote)


class NvimComponent(Generic[R], Logging):

    def __init__(self, vim: Nvim, target: Union['NvimComponent', 'AsyncVimProxy', R], prefix: str) -> None:
        self.vim = vim
        self.target = target
        self.prefix = prefix
        if ribosome.in_vim and isinstance(target, (AsyncVimProxy, NvimComponent)):
            msg = '{} created with non-native target {}'
            raise Exception(msg.format(self, target))
        self._vars = Vars(self)
        self._options = Options(self)

    def __repr__(self) -> str:
        n = '' if not_on_main_thread() else self._details
        return '{}({}, {})'.format(self.__class__.__name__, self.prefix, n)

    @property
    def vars(self) -> 'AsyncVimProxy[Vars]':
        return AsyncVimProxy(self._vars, self)

    @property
    def options(self) -> 'AsyncVimProxy[Options]':
        return AsyncVimProxy(self._options, self)

    @property
    def _details(self) -> str:
        return ''

    @property
    def loop(self) -> Any:
        return self.vim._session._async_session._msgpack_stream._event_loop._loop

    def delay(self, f: Callable[[], None], timeout: float) -> None:
        cb = lambda: self._run_on_main_thread(f)
        threading.Timer(1.0, cb).start()

    def async_call(self, f: Callable[..., Any], *a: Any, **kw: Any) -> Any:
        ''' run a callback function on the main thread and return its
        value (blocking). the callback receives 'self' as an argument.
        '''
        if not shutdown:
            msg = 'running {} on main thread blocking'
            self.log.debug2(lambda: msg.format(format_funcall(f.__name__, a, kw)))
            result_fut = futures.Future()  # type: futures.Future
            @functools.wraps(f)
            def cb() -> None:
                result_fut.set_result(f(self, *a, **kw))
            self._run_on_main_thread(cb)
            result = result_fut.result()
            self.log.debug2(lambda: f'async returns {result}')
            return result
        else:
            return f(self, *a, **kw)

    async = async_call

    def _run_on_main_thread(self, f: Callable[..., Any], *a: Any, **kw: Any) -> None:
        ''' run a callback function on the host's main thread
        '''
        frame = inspect.currentframe()
        def dispatch() -> None:
            self.log.debug2(lambda: 'running on main thread: {}'.format(format_funcall(f.__name__, a, kw)))
            try:
                f(*a, **kw)
            except NvimError as e:
                self._report_nvim_error(e, frame)
            else:
                self.log.debug2('{} successful'.format(f.__name__))
        self.vim._session.threadsafe_call(dispatch)

    def _report_nvim_error(self, err: str, frame: traceback.FrameSummary) -> None:
        self.log.error('async vim call failed with \'{}\''.format(decode(err)))
        self.log.error(''.join(traceback.format_stack(frame)[:-5]))

    @contextmanager
    def main_event_loop(self):
        ''' run the asyncio main loop on vim's pyuv main thread and
        yield the loop to the with statement in the current thread.
        '''
        def get_event_loop(v) -> Any:
            return asyncio.get_event_loop()
        main = self.async_call(get_event_loop)
        fut = main.create_future()
        if not main.is_running():
            self._run_on_main_thread(main.run_until_complete, fut)
            start = time.time()
            while not main.is_running() and time.time() - start < 1:
                time.sleep(0.01)
            if not main.is_running():
                self.log.debug('timed out waiting for main loop to start')
        try:
            yield main
        except:
            self.log.exception('error while running callback in main loop:')
        main.call_soon_threadsafe(L(fut.set_result)(True))

    @contextmanager
    def threadsafe_subprocess(self):
        def set_watcher(vim: Any) -> None:
            try:
                asyncio.get_child_watcher().attach_loop(asyncio.get_event_loop())
            except Exception:
                pass
        self.async_call(set_watcher)
        with self.main_event_loop() as main:
            yield main

    @property
    def proxy(self):
        return AsyncVimProxy(self, self)

    def prefixed(self, name: str):
        return '{}_{}'.format(self.prefix, name) if self.prefix else name

    def clean(self):
        self.vars.clean()

    def _cmd(self, line: str, verbose=False, sync=False) -> Either[Exception, None]:
        l = line if verbose else 'silent {}'.format(line)
        return Try(self.vim.command, l, async=not sync)

    def cmd(self, line: str, verbose=False, sync=False, report=True) -> Either[Exception, None]:
        ''' Wrap **Nvim.command**, default to async.
        **verbose** prevents the use of **silent**, which is used by
        default because headless nvim will deadlock if a command's
        output requires user input to proceed (e.g. multiline output)
        '''
        return (
            self._cmd(line, verbose, sync)
            .bieffect(
                L(self._cmd_error)(line, _, sync, report),
                L(self._cmd_success)(line, _, sync)
            )
        )

    def cmd_sync(self, line: str, verbose=False, report=True) -> Either[Exception, None]:
        return self.cmd(line, verbose=verbose, sync=True, report=report)

    def cmd_output(self, line: str) -> List[str]:
        return Lists.wrap(self.vim.command_output('silent {}'.format(line)).splitlines())

    def _cmd_success(self, line: str, result, sync: bool) -> None:
        if sync:
            self.log.debug1(lambda: f'result of cmd \'{line}\': {result}')
        else:
            self.log.debug1('async cmd \'{}\''.format(line))

    def _cmd_error(self, line, exc, sync, report: bool) -> None:
        if not report:
            return
        err = exc.cause if isinstance(exc, IOException) else exc
        if amino.development:
            a = '' if sync else 'a'
            msg = f'running nvim {a}sync cmd `{line}`'
            if isinstance(err, Exception):
                self.log.caught_exception_error(msg, err)
            else:
                self.log.error(f'{msg}: {err}')
        else:
            self.log.error(decode(err))

    def vcmd(self, line: str, sync=False, report=True) -> Either[Exception, None]:
        return self.cmd(line, verbose=True, sync=sync, report=report)

    def normal(self, keys) -> Either[Exception, None]:
        return self.cmd_sync('normal! {}'.format(keys))

    @property
    def syntax(self) -> 'Syntax':
        return Syntax(self)

    @lazy
    def root(self) -> 'NvimFacade':
        return NvimFacade(self.vim, self.prefix)


class NvimCmd:

    def __init__(self, vim: NvimComponent, name: str, args: str='', range=Empty(), silent=True) -> None:
        self.vim = vim
        self.name = name
        self.args = args
        self.range = range
        self.silent = silent

    def _cmdline(self, silent=None):
        sil = self.silent if silent is None else silent
        s = 'silent! ' if sil else ''
        r = self.range | ''
        return '{}{}{} {}'.format(s, r, self.name, self.args)

    def run_sync(self, silent=None):
        return self.vim.cmd_sync(self._cmdline(silent))

    def run_async(self, silent=None):
        return self.vim.cmd(self._cmdline(silent))

    def run_output(self, silent=None):
        return self.vim.cmd_output(self._cmdline(silent))


class HasBuffer(NvimComponent, metaclass=abc.ABCMeta):

    @property
    def buffer(self):
        return Buffer(self.vim, self._internal_buffer, self.prefix).proxy

    @abc.abstractproperty
    def _internal_buffer(self):
        ...

    def __eq__(self, other):
        if hasattr(other, '_internal_buffer'):
            return self._internal_buffer == other._internal_buffer


class HasBuffers(HasBuffer):

    @property
    def buffers(self):
        return List(*self._internal_buffers)\
            .map(lambda a: Buffer(self.vim, a, self.prefix).proxy)

    @property
    def _internal_buffers(self):
        return self.vim.buffers


class HasWindow(HasBuffer):

    @property
    def window(self):
        return Window(self.vim, self._internal_window, self.prefix).proxy

    @abc.abstractproperty
    def _internal_window(self):
        ...

    @property
    def _internal_buffer(self):
        return self.window._internal_buffer

    def vsplit(self):
        self.cmd('vsplit')
        return self.window

    def new(self):
        self.cmd_sync('new')
        return self.window

    def vnew(self):
        self.cmd_sync('vnew')
        return self.window


class HasWindows(HasBuffers, HasWindow):

    @property
    def windows(self):
        return (
            self._internal_windows /
            (lambda a: Window(self.vim, a, self.prefix).proxy)
        )

    @property
    def _internal_windows(self):
        return List.wrap(self.vim.windows)

    def focus(self, number: int):
        return NvimCmd(self, 'wincmd', 'w', range=Just(number))


class HasTab(HasWindow):

    @property
    def tab(self):
        return Tab(self.vim, self._internal_tab, self.prefix).proxy

    @abc.abstractproperty
    def _internal_tab(self):
        ...

    @property
    def _internal_window(self):
        return self.tab._internal_window


class HasTabs(HasTab):

    @property
    def tabs(self):
        return List(*self._internal_tabs)\
            .map(lambda a: Tab(self.vim, a, self.prefix).proxy)

    @property
    def _internal_tabs(self):
        return self.vim.tabpages

    def tabnew(self):
        self.cmd('tabnew')
        return self.tab


class Buffer(HasWindow):

    @property
    def _details(self):
        return Try(lambda: self.desc) | ''

    @property
    def _internal_buffer(self):
        return self.target

    @property
    def _internal_window(self):
        return self.window.target

    @property
    def modified(self):
        return self.options('modified').contains(True)

    @property
    def name(self):
        return self._internal_buffer.name

    @property
    def content(self):
        return List.wrap(self.target[:]).map(decode)

    def set_content(self, text, rng=slice(None)):
        self.target[rng] = text

    def nmap(self, keyseq, dispatch):
        return self.cmd('nmap <buffer><silent> {} {}'.format(keyseq, dispatch))

    def set_modifiable(self, value):
        self.options.set_b('modifiable', value)

    @property
    def modifiable(self):
        return self.options.b('modifiable')

    @property
    def desc(self):
        n = self.name
        name = ' {}'.format(n) if n else ''
        return 'buf #{}{}'.format(self._internal_buffer.number, name)

    def autocmd(self, name, cmd):
        return NvimCmd(self, 'autocmd', '{} <buffer> {}'.format(name, cmd))

    def pvar_or_global(self, name):
        return self.vars.p(name).or_else(lambda: self.root.vars.p(name))

    @property
    def id(self):
        return self._internal_buffer.number

    @property
    def line_count(self):
        return self.content.length


class Window(HasTab, HasBuffers):

    @property
    def _details(self):
        return Try(lambda: self.buffer.desc) | ''

    @property
    def _internal_buffer(self):
        return self.target.buffer

    @property
    def _internal_tab(self):
        return self.tabpage

    @property
    def _internal_window(self):
        return self.target

    def reload(self):
        self.focus()
        if not self.buffer.modified:
            self.cmd('silent edit')

    # FIXME not constrained to visible tab
    def focus(self):
        self.vim.current.window = self.target

    @property
    def cursor(self):
        return List.wrap(self.target.cursor)

    @property
    def line(self):
        return self.cursor.head

    @property
    def line0(self):
        return self.line / (_ - 1)

    def close(self):
        self.cmd('close')

    def close_unless_last(self):
        if self.tab.windows.length > 1:
            self.close()

    def set_cursor(self, line, col=0):
        buf_max = self.buffer.content.length
        self.target.cursor = (min(line, buf_max), col)

    @property
    def width(self):
        return parse_int(self.target.width)

    @property
    def height(self):
        return parse_int(self.target.height)

    @property
    def size(self):
        return self.width & self.height

    def visual_line(self, start, end):
        self.set_cursor(start)
        self.normal('V')
        self.set_cursor(end)


class Tab(HasWindows):

    @property
    def _internal_window(self):
        return self.target.window

    def close(self):
        self.cmd('tabclose')


class NvimFacade(HasTabs, HasWindows, HasBuffers, HasTab):

    @staticmethod
    def stdio_with_logging(name: str) -> 'NvimFacade':
        from ribosome.logging import nvim_logging
        native = neovim.attach('stdio')
        vim = NvimFacade(native, name)
        nvim_logging(vim)
        return vim

    def __init__(self, vim: neovim.Nvim, prefix: str) -> None:
        super().__init__(vim, vim, prefix)

    @property
    def channel_id(self) -> int:
        return self.vim.channel_id

    def runtime(self, path: str, verbose=True):
        return self.cmd_sync('runtime! {}.vim'.format(path), verbose=verbose)

    def _echo(self, lines: List[str]) -> None:
        return lines.traverse(L(self._cmd)(_, True), Right)

    def echo(self, text: Union[str, List[str]]):
        self._echo(echo(text, 'echo', prefix=Just(self.prefix)))

    def echom(self, text: Union[str, List[str]]):
        self._echo(echo(text, prefix=Just(self.prefix)))

    def echohl(self, hl: Union[str, List[str]], text: Union[str, List[str]]):
        self._echo(echohl(text, hl, prefix=Just(self.prefix)))

    def echowarn(self, text: Union[str, List[str]]):
        self.echohl('WarningMsg', text)

    def echoerr(self, text: Union[str, List[str]]):
        self.echohl('ErrorMsg', text)

    def multi_line_info(self, text: List[str]):
        ''' as a workaround for possible bug in embedded nvim,
        only write in bulk when in production
        '''
        if amino.integration_test:
            text.foreach(self.log.info)
        else:
            self.log.info('\n'.join(text))

    def doautocmd(self, name, pat='') -> Either[Exception, None]:
        c = 'doautocmd <nomodeline> {} {}'.format(name, pat)
        return self.cmd(c)

    def uautocmd(self, name) -> Either[Exception, None]:
        return self.doautocmd('User', name)

    def pautocmd(self, name) -> Either[Exception, None]:
        return self.uautocmd('{}{}'.format(camelcaseify(self.prefix), name))

    def autocmd(self, name, pattern, cmd) -> NvimCmd:
        return NvimCmd(self, 'autocmd', '{} {} {}'.format(name, pattern, cmd))

    def _call(self, name: str, *args: Any, sync=True, **kw: Any) -> Either[Exception, A]:
        return Try(self.vim.call, name, *args, async=not sync, **kw)

    def call(self, name: str, *args: Any, sync=True, **kw: Any) -> Either[str, A]:
        return (
            self._call(name, *args, sync=sync, **kw)
            .leffect(L(self._call_error)(name, args, kw, _, sync))
            .lmap(lambda a: 'vim call `{}` failed: {}'.format(format_funcall(name, args, kw), a))
        )

    def _call_error(self, name: str, args: tuple, kw: dict, err: Exception, sync: bool) -> None:
        a = '' if sync else 'a'
        header = f'calling {a}sync nvim function `{format_funcall(name, args, kw)}`'
        thread_error = str(err) == "'NoneType' object has no attribute 'switch'"
        msg = 'thread scheduling error' if isinstance(err, Exception) and thread_error else str(err)
        if isinstance(err, Exception) and amino.development:
            self.log.caught_exception(header, err)
        self.log.error(f'{header}: {msg}')

    def eval(self, expr):
        return self.vim.eval(expr)

    @property
    def current(self):
        return self.vim.current

    @property
    def _internal_tab(self):
        return self.current.tabpage

    @property
    def _internal_window(self):
        return self.current.window

    @property
    def _internal_buffer(self):
        return self.current.buffer

    @property
    def modified(self):
        return self.buffer.modified

    def reload_window(self):
        self.window.reload()

    def reload_windows(self):
        self.tab.windows.foreach(__.reload())

    def cursor(self, line, col):
        return self.call('cursor', line, col)

    def feedkeys(self, keyseq, *a, **kw):
        self.vim.feedkeys(keyseq, *a, **kw)

    def edit(self, path):
        if self.buffer.modifiable:
            self.write()
        return NvimCmd(self, 'edit!', str(path))

    def write(self):
        return NvimCmd(self, 'write!')

    @property
    def messages(self):
        return self.cmd_output('messages')

    def import_pvar_path(self, name: str) -> Maybe[A]:
        return self.buffer.vars.p(name).o(lambda: self.vars.p(name)).flat_map(Either.import_path)

    def function_exists(self, name: str) -> Boolean:
        return self.call('exists', f'*{name}').true

    def command_exists(self, name: str) -> Boolean:
        return self.call('exists', f':{name}').contains(2)

    def execute(self, code: Union[str, List[str]]) -> None:
        lines = code if isinstance(code, List) else Lists.lines(code)
        return self.call('execute', list(lines))

    def run_once_defined(self, job: Callable[[], Either[str, A]], err: Callable[[str], str], timeout: int=3
                         ) -> Either[str, A]:
        result: Either[str, A] = Left('not started yet')
        start = time.time()
        while not result.is_right and time.time() - start < timeout:
            time.sleep(.01)
            result = job()
        return result.lmap(err)

    def call_once_defined(self, name: str, *args: str, timeout: int=3, **kw: Any) -> Either[str, A]:
        return self.run_once_defined(L(self._call)(name, *args, **kw), lambda a: f'function {name} did not appear: {a}',
                                     timeout=timeout)

    def cmd_once_defined(self, name: str, *args: str, timeout: int=3, **kw: Any) -> Either[str, A]:
        return self.run_once_defined(
            L(self.cmd_sync)(name, *args, report=False, **kw),
            lambda a: f'cmd {name} did not appear: {a}',
            timeout=timeout
        )

    def wait_for_command(self, name: str, timeout: int=3) -> Either[str, None]:
        f = lambda: self.command_exists(name).e('', None)
        return self.run_once_defined(f, lambda a: 'cmd {name} did not appear', timeout=timeout)

    def define_function(self, name: str, params: List[str], body: str) -> None:
        self.cmd_sync(f'function! {name}({params.join_comma})\n{body}\nendfunction')

    def cd(self, dir: Path) -> None:
        return self.vim.chdir(str(dir))

    @property
    def cwd(self) -> Either[str, Path]:
        return self.call('getcwd') // L(Try)(Path, _)

    @property
    def flags(self) -> 'Flags':
        return Flags(self.vars, False)

    @property
    def pflags(self) -> 'Flags':
        return Flags(self.vars, True)

class AsyncVimCallProxy(Generic[A]):

    def __init__(self, target: A, vim: NvimComponent, name: str) -> None:
        self._target = target
        self._vim = vim
        self.name = name

    def __call__(self, *a: Any, **kw: Any) -> Any:
        def proxy_call(vim: NvimComponent, target: A, name: str, *a: Any, **kw: Any) -> Any:
            return getattr(target, name)(*a, **kw)
        return self._vim.async_call(proxy_call, self._target, self.name, *a, **kw)

    def __repr__(self) -> str:
        return '{}({}, {})'.format(self.__class__.__name__, self.name, self._target)


class AsyncVimProxy(Generic[A]):
    allow_async_relay = True

    def __init__(self, target: A, vim: NvimComponent) -> None:
        self._vim = vim
        self._target = target
        self._target_tpe = type(target)

    @property
    def _need_relay(self):
        return self.allow_async_relay and not_on_main_thread()

    def __getattr__(self, name):
        if self._need_relay:
            return self.async_relay(name)
        else:
            return getattr(self._target, name)

    def async_relay(self, name):
        if (hasattr(self._target_tpe, name)):
            attr = getattr(self._target_tpe, name)
            if isinstance(attr, FunctionType):
                return AsyncVimCallProxy(self._target, self._vim, name)
            else:
                return self._async_attr(name)
        elif hasattr(self._target, name):
            return self._async_attr(name)
        else:
            return getattr(self._target, name)

    def _async_attr(self, name: str) -> Any:
        def proxy_getattr(vim, target: NvimComponent, name: str) -> Any:
            return getattr(target, name)
        return self._vim.async_call(proxy_getattr, self._target, name)

    def __eq__(self, other):
        return self.__getattr__('__eq__')(other)

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self._target)

    def __call__(self, *a, **kw):
        return self.__getattr__('__call__')(*a, **kw)


class OptVar(Logging, abc.ABC):

    @abc.abstractmethod
    def _get(self, name: str) -> Optional[Any]:
        ...

    @abc.abstractmethod
    def _set(self, name: str, value: A) -> None:
        ...

    @property
    def _desc(self) -> str:
        ...

    def __call__(self, name: str) -> Either[str, Any]:
        v = self._get(name)
        if v is None:
            msg = '{} not found: {}'.format(self._desc, name)
            self.log.debug(msg)
            return Left(msg)
        else:
            return Right(decode(v))

    def set(self, name: str, value: A) -> None:
        self.log.debug('setting {} {} to \'{}\''.format(self._desc, name, value))
        try:
            self._set(name, value)
        except:
            msg = 'setting vim {} {} to {}'
            self.log.exception(msg.format(self._desc, name, value))

    def path(self, name: str) -> Either[str, Path]:
        return self(name).map(lambda a: Path(a).expanduser())

    def dir(self, name: str) -> Either[str, Path]:
        var = self.path(name)
        val = var.filter(__.is_dir())
        if not val.present:
            msg = 'g:{} is not a directory ({})'
            self.log.error(msg.format(name, var))
        return val

    def typed(self, tpe: Type[A], value: Either[str, Any]) -> Either[str, A]:
        def check(v: A) -> Either[str, A]:
            if isinstance(v, tpe):
                return Right(v)
            else:
                msg = 'invalid type {} for {} {} (wanted {})'.format(type(v), self._desc, v, tpe)
                self.log.error(msg)
                return Left(msg)
        return value.flat_map(check)

    def s(self, name: str) -> Either[str, str]:
        return self.typed(str, self(name))

    def l(self, name: str) -> Either[str, List]:
        return self.typed(list, self(name)) / Lists.wrap

    def d(self, name: str) -> Either[str, dict]:
        return self.typed(dict, self(name))

    def m(self, name: str) -> Either[str, Map]:
        return self.d(name) / Map

    def b(self, name: str) -> Either[str, Boolean]:
        return self.typed(bool, self(name)) / Boolean

    def i(self, name: str) -> Either[str, int]:
        return self.typed(int, self(name))

    def exists(self, name: str) -> Boolean:
        return self(name).is_right


# FIXME recording variable name for eventual cleanup only works with a global
# NvimFacade instance
class Vars(OptVar):

    def __init__(self, vim: NvimComponent) -> None:
        self.vim = vim
        self._vars: set = set()

    def __repr__(self) -> str:
        return '{}({!r})'.format(self.__class__.__name__, self.vim)

    @property
    def internal(self):
        return self.vim.target.vars

    def _get(self, name):
        return self.internal.get(name)

    @property
    def _desc(self):
        return 'variable'

    def prefixed(self, name):
        return self.vim.prefixed(name)

    def _set(self, name, value):
        self.internal[name] = value
        self._vars.add(name)

    def clean(self):
        for name in self._vars:
            if not re.match('^[A-Z].*', name):
                del self.internal[name]
        self._vars = set()

    def p(self, name) -> Either[str, str]:
        return self(self.prefixed(name))

    def set_p(self, name, value):
        self.set(self.prefixed(name), value)

    def ppath(self, name: str) -> Either[str, Path]:
        return self.path(self.prefixed(name))

    def ppathl(self, name: str) -> Either[str, Path]:
        return self.pl(name)\
            .map(lambda l: l.map(Path).map(__.expanduser()))

    def pdir(self, name: str) -> Either[str, Path]:
        return self.dir(self.prefixed(name))

    def ps(self, name):
        return self.typed(str, self.vars.p(name))

    def pl(self, name):
        return self.typed(list, self.p(name))

    def pd(self, name):
        return self.typed(dict, self.p(name))

    def pm(self, name):
        return self.pd(name) / Map

    def pb(self, name: str) -> Either[str, Boolean]:
        return self.typed(bool, self.p(name)) / Boolean

    def pi(self, name: str):
        return self.typed(int, self.p(name))

    @property
    def all(self):
        rex = re.compile('^#(?P<num>\d+)$')
        def parse(k, v):
            v2 = (Maybe(rex.match(v)) / __.groupdict() / Map //
                  __.get('num') // parse_int)
            return k, (v2 | v)
        lines = (
            (self.vim.cmd_output('silent let') / __.split(maxsplit=1) /
                List.wrap)
            .filter(_.length == 2)
            .filter_not(lambda a: a[0].startswith('v:'))
            .map2(parse)
        )
        return Map(lines)

    def ensure(self, name: str, default: Any) -> None:
        if not self.exists(name):
            self.set(name, default)

    def ensure_p(self, name: str, default: Any) -> None:
        self.ensure(self.prefixed(name), default)


class Options(OptVar):

    def __init__(self, vim) -> None:
        self.vim = vim

    def _get(self, name):
        return self.vim.target.options.get(name)

    @property
    def _desc(self):
        return 'option'

    def _set(self, name: str, value):
        self.vim.target.options[name] = value

    def s(self, name: str):
        return self.typed(str, self(name))

    def set_s(self, name: str, value):
        return self.set(name, str(value))

    def l(self, name: str) -> List[str]:
        return self.s(name)\
            .map(lambda a: a.split(','))\
            .map(List.wrap)\
            .get_or_else(List())

    def amend_l(self, name: str, value):
        if not isinstance(value, list):
            value = List(value)
        new_value = (self.l(name) + list(map(str, value))).distinct
        self.set_l(name, new_value)

    def set_l(self, name: str, value: List[str]):
        self.vim.target.options[name] = ','.join(value)

    def b(self, name: str):
        return self.typed(bool, self(name))

    def set_b(self, name: str, value):
        return self.set(name, bool(value))


class Flags:

    def __init__(self, vars, prefix: bool) -> None:
        self.vars = vars
        self.prefix = prefix

    def get(self, name, default=False):
        v = (self.vars.p(name) if self.prefix else self.vars(name))\
            .get_or_else(default)
        return Boolean(v)

    def __getattr__(self, name):
        return self.get(name)


class HasNvim:

    def __init__(self, vim: NvimFacade) -> None:
        self.vim = vim
        self.flags = Flags(vim.vars, False)
        self.pflags = Flags(vim.vars, True)


class Syntax(Logging):

    def __init__(self, target):
        self.target = target

    def _opts(self, *a, **kw):
        o = List.wrap(a) + Map(kw).to_list.map2('{}={}'.format)
        return ' '.join(o)

    def match(self, group, pat, *a, **kw):
        return self.cmd('match', group, pat, *a, **kw)

    def cmd(self, cmdname, group, pat, *a, **kw):
        c = 'syntax {} {} /{}/ {}'.format(cmdname, group, pat, self._opts(*a, **kw))
        self.target.cmd(c)

    def link(self, group, to):
        c = 'highlight link {} {}'.format(group, to)
        self.target.cmd(c)

    def highlight(self, group, *a, **kw):
        c = 'highlight {} {}'.format(group, self._opts(*a, **kw))
        self.target.cmd(c)

__all__ = ('NvimFacade', 'HasNvim')



import abc
import inspect
from typing import Callable, Type, TypeVar, Generic, Iterable, cast, Union, Any, Optional

from amino.options import env_xdg_data_dir
from amino import List, Either, Lists, Map, _, L, __, Path, Right, Try, Nil, Just
from amino.func import flip
from amino.util.string import ToStr, snake_case

from ribosome.nvim import NvimIO, NvimFacade
from ribosome.nvim.components import NvimComponent
from ribosome.logging import Logging
from ribosome.data import Data
from ribosome.machine.message_base import Message
from ribosome.request.command import msg_command, json_msg_command
from ribosome.request.function import msg_function
from ribosome.request.autocmd import msg_autocmd
from ribosome.record import field


A = TypeVar('A', contravariant=True)
B = TypeVar('B')


class PluginSetting(Generic[A, B], ToStr):

    def __init__(
            self,
            name: str,
            desc: str,
            help: str,
            prefix: bool,
            tpe: Type[A],
            ctor: Callable[[A], B],
            default: B,
    ) -> None:
        self.name = name
        self.desc = desc
        self.help = help
        self.prefix = prefix
        self.tpe = tpe
        self.ctor = ctor
        self.default = default

    @property
    def value(self) -> NvimIO[Either[str, A]]:
        def read(v: NvimComponent) -> Either[str, A]:
            vars = v.vars
            getter = vars.p if self.prefix else vars
            return vars.typed(self.tpe, getter(self.name)) // self.ctor
        return NvimIO(read)

    @property
    def value_or_default(self) -> NvimIO[A]:
        return self.value / __.get_or_else(self.default)

    def value_or(self, default: A) -> NvimIO[A]:
        return self.value / __.get_or_else(default)

    def _arg_desc(self) -> List[str]:
        return List(self.name, str(self.prefix), str(self.tpe))


def setting_ctor(tpe: Type[A], ctor: Callable[[A], B]) -> Callable[[str, str, str, bool, B], PluginSetting[A, B]]:
    def setting(name: str, desc: str, help: str, prefix: bool, default: B) -> PluginSetting[A, B]:
        return PluginSetting(name, desc, help, prefix, tpe, ctor, default)
    return setting


state_dir_help_default = '''This directory is used to persist the plugin's current state.'''
state_dir_base_default = env_xdg_data_dir.value / Path | (Path.home() / '.local' / 'share')


class PluginSettings(ToStr, Logging):

    def __init__(self, name: str, state_dir_help: str=state_dir_help_default) -> None:
        self.name = name
        self.components = list_setting('components', 'names or paths of active components', components_help, True, Nil)
        self.state_dir = path_setting('state_dir', 'state persistence directory', state_dir_help, True,
                                      state_dir_base_default / self.name)

    def all(self) -> Map[str, PluginSetting]:
        settings = inspect.getmembers(self, L(isinstance)(_, PluginSetting))
        return Map(Lists.wrap(settings).map(_[1]).apzip(_.name).map2(flip))

    def _arg_desc(self) -> List[str]:
        return List(str(self.all()))


def path_list(data: list) -> Either[str, List[Path]]:
    return Lists.wrap(data).traverse(lambda a: Try(Path, a) / __.expanduser(), Either)


components_help = '''The plugin can run an arbitrary set of sub-components that inherit the class `Component`.
They receive all messages the core of the plugin processes and have access to the plugin state.
They can define nvim request handlers with the decorator `trans.xyz`.
The entries of this setting can either be names of the builtin components or arbitrary python module paths that define
custom components.
'''
str_setting = setting_ctor(str, lambda a: Right(a))
list_setting = setting_ctor(list, cast(Callable[[Iterable[A]], Either[str, List[B]]], (lambda a: Right(Lists.wrap(a)))))
path_setting = setting_ctor(str, (lambda a: Try(Path, a)))
path_list_setting = setting_ctor(list, path_list)
map_setting = setting_ctor(dict, lambda a: Right(Map(a)))
path_map_setting = setting_ctor(dict, lambda a: Try(Map, a).valmap(lambda b: Path(b).expanduser()))

M = TypeVar('M', bound=Message)


class RequestDispatcher(ToStr):

    @abc.abstractmethod
    def decorator(self) -> Callable[..., Callable]:
        ...

    @abc.abstractproperty
    def args(self) -> List[Any]:
        ...

    @abc.abstractproperty
    def name(self) -> str:
        ...


class MsgDispatcher(Generic[M], RequestDispatcher):

    def __init__(self, msg: Type[M]) -> None:
        self.msg = msg

    def _arg_desc(self) -> List[str]:
        return List(str(self.msg))

    @property
    def args(self) -> List[Any]:
        return List(self.msg)

    @property
    def name(self) -> str:
        return snake_case(self.msg.__name__)


class MsgCmd(Generic[M], MsgDispatcher[M]):

    def decorator(self) -> Callable[..., Callable]:
        return msg_command


class JsonMsgCmd(Generic[M], MsgDispatcher[M]):

    def decorator(self) -> Callable[..., Callable]:
        return json_msg_command


class MsgFun(Generic[M], MsgDispatcher[M]):

    def decorator(self) -> Callable[..., Callable]:
        return msg_function


class MsgAutocmd(Generic[M], MsgDispatcher[M]):

    def decorator(self) -> Callable[..., Callable]:
        return msg_autocmd


class PrefixStyle(ToStr):

    def _arg_desc(self) -> List[str]:
        return Nil


class Short(PrefixStyle):
    pass


class Full(PrefixStyle):
    pass


class Plain(PrefixStyle):
    pass


class RequestHandler(ToStr):

    def __init__(
            self,
            dispatcher: RequestDispatcher,
            name: str,
            prefix: PrefixStyle=Short(),
            options: Map[str, Any]=Map()
    ) -> None:
        self.dispatcher = dispatcher
        self.name = name
        self.prefix = prefix
        self.options = options

    @staticmethod
    def msg_cmd(msg: Type[M]) -> 'RequestHandlerBuilder':
        return RequestHandlerBuilder(MsgCmd(msg))

    @staticmethod
    def msg_fun(msg: Type[M]) -> 'RequestHandlerBuilder':
        return RequestHandlerBuilder(MsgFun(msg))

    @staticmethod
    def msg_autocmd(msg: Type[M]) -> 'RequestHandlerBuilder':
        return RequestHandlerBuilder(MsgAutocmd(msg))

    @staticmethod
    def json_msg_cmd(msg: Type[M]) -> 'RequestHandlerBuilder':
        return RequestHandlerBuilder(JsonMsgCmd(msg))

    def _arg_desc(self) -> List[str]:
        return List(str(self.dispatcher), self.name, str(self.prefix), str(self.options))


class RequestHandlerBuilder:

    def __init__(self, dispatcher: RequestDispatcher) -> None:
        self.dispatcher = dispatcher

    def __call__(self, name: str=None, prefix: PrefixStyle=Short(), **options: Any) -> RequestHandler:
        name1 = name or self.dispatcher.name
        return RequestHandler(self.dispatcher, name1, prefix, Map(options))


class RequestHandlers(ToStr):

    @staticmethod
    def cons(*handlers: RequestHandler) -> 'RequestHandlers':
        return RequestHandlers(Map(Lists.wrap(handlers).apzip(_.name).map2(flip)))

    def __init__(self, handlers: Map[str, RequestHandler]) -> None:
        self.handlers = handlers

    def _arg_desc(self) -> List[str]:
        return List(str(self.handlers))


Settings = TypeVar('Settings', bound=PluginSettings)
S = TypeVar('S', bound='AutoData')


class Config(Generic[Settings, S], ToStr):

    @staticmethod
    def from_opt(data: Map) -> 'Config':
        return Config(data.lift('name') | 'no name in json', data.lift('prefix') | None)

    def __init__(
            self,
            name: str,
            prefix: Optional[str]=None,
            components: Map[str, Union[str, type]]=Map(),
            state_type: Optional[Type[S]]=None,
            state_ctor: Optional[Callable[['Config', NvimFacade], S]]=None,
            settings: Optional[Settings]=None,
            request_handlers: List[RequestHandler]=Nil,
            core_components: List[str]=Nil,
            default_components: List[str]=Nil
    ) -> None:
        self.name = name
        self.prefix = prefix or name
        self.components = components
        self.state_type = state_type or AutoData
        self.state_ctor = state_ctor or (lambda c, v: self.state_type(config=c, vim_facade=Just(v)))
        self.settings = settings or PluginSettings(name=name)
        self.request_handlers = RequestHandlers.cons(*request_handlers)
        self.core_components = core_components
        self.default_components = default_components

    def _arg_desc(self) -> List[str]:
        return List(str(self.components), str(self.settings), str(self.request_handlers))

    def state(self, vim: NvimFacade) -> S:
        return self.state_ctor(self, vim)

    @property
    def json_repr(self) -> dict:
        return dict(__type__='ribosome.settings.Config', name=self.name, prefix=self.prefix)


class AutoData(Data):
    config: Config = field(Config)

    @property
    def _str_extra(self) -> List[Any]:
        return List(self.config)

    @property
    def settings(self) -> Settings:
        return self.config.settings

__all__ = ('PluginSetting', 'setting_ctor', 'PluginSettings', 'str_setting', 'list_setting', 'Config',
           'RequestHandlers', 'RequestHandler')
